(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var global = Package.meteor.global;
var meteorEnv = Package.meteor.meteorEnv;
var ECMAScript = Package.ecmascript.ECMAScript;
var Accounts = Package['accounts-base'].Accounts;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormPermissions = Package['sandstorm-permissions'].SandstormPermissions;
var meteorInstall = Package.modules.meteorInstall;
var Buffer = Package.modules.Buffer;
var process = Package.modules.process;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var meteorBabelHelpers = Package['babel-runtime'].meteorBabelHelpers;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var inMeteor, promiseToFuture, waitPromise, SANDSTORM_ALTHOME, SandstormBackend;

var require = meteorInstall({"node_modules":{"meteor":{"sandstorm-backend":{"sandstorm-backend.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-backend/sandstorm-backend.js                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               //
// All rights reserved.                                                                                                //
//                                                                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");                                                     //
// you may not use this file except in compliance with the License.                                                    //
// You may obtain a copy of the License at                                                                             //
//                                                                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        //
//                                                                                                                     //
// Unless required by applicable law or agreed to in writing, software                                                 //
// distributed under the License is distributed on an "AS IS" BASIS,                                                   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            //
// See the License for the specific language governing permissions and                                                 //
// limitations under the License.                                                                                      //
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 17
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 18
var Crypto = Npm.require("crypto");                                                                                    // 19
var Future = Npm.require("fibers/future");                                                                             // 20
var Promise = Npm.require("es6-promise").Promise;                                                                      // 21
                                                                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 23
  callback();                                                                                                          // 24
});                                                                                                                    // 25
                                                                                                                       //
inMeteor = function inMeteor(callback) {                                                                               // 27
  // Calls the callback in a Meteor context.  Returns a Promise for its result.                                        //
  return new Promise(function (resolve, reject) {                                                                      // 29
    inMeteorInternal(function () {                                                                                     // 30
      try {                                                                                                            // 31
        resolve(callback());                                                                                           // 32
      } catch (err) {                                                                                                  // 33
        reject(err);                                                                                                   // 34
      }                                                                                                                // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
};                                                                                                                     // 38
                                                                                                                       //
promiseToFuture = function promiseToFuture(promise) {                                                                  // 40
  var result = new Future();                                                                                           // 41
  promise.then(result["return"].bind(result), result["throw"].bind(result));                                           // 42
  return result;                                                                                                       // 43
};                                                                                                                     // 44
                                                                                                                       //
waitPromise = function waitPromise(promise) {                                                                          // 46
  return promiseToFuture(promise).wait();                                                                              // 47
};                                                                                                                     // 48
                                                                                                                       //
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                           // 50
                                                                                                                       //
SandstormBackend = function SandstormBackend(db, backendCap) {                                                         // 52
  this._db = db;                                                                                                       // 53
  this._backendCap = backendCap;                                                                                       // 54
};                                                                                                                     // 55
                                                                                                                       //
SandstormBackend.prototype.cap = function () {                                                                         // 57
  return this._backendCap;                                                                                             // 58
};                                                                                                                     // 59
                                                                                                                       //
SandstormBackend.prototype.deleteUser = function (userId) {                                                            // 61
  return waitPromise(this._backendCap.deleteUser());                                                                   // 62
};                                                                                                                     // 63
                                                                                                                       //
SandstormBackend.prototype.shutdownGrain = function (grainId, ownerId, keepSessions) {                                 // 65
  if (!keepSessions) {                                                                                                 // 66
    Sessions.remove({ grainId: grainId });                                                                             // 67
  }                                                                                                                    // 68
                                                                                                                       //
  var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                                  // 70
  return grain.shutdown().then(function () {                                                                           // 71
    grain.close();                                                                                                     // 72
    throw new Error("expected shutdown() to throw disconnected");                                                      // 73
  }, function (err) {                                                                                                  // 74
                                                                                                                       //
    grain.close();                                                                                                     // 76
    if (err.kjType !== "disconnected") {                                                                               // 77
      throw err;                                                                                                       // 78
    }                                                                                                                  // 79
  });                                                                                                                  // 80
};                                                                                                                     // 81
                                                                                                                       //
SandstormBackend.prototype.deleteGrain = function (grainId, ownerId) {                                                 // 83
  // We leave it up to the caller if they want to actually wait, but some don't so we report                           //
  // exceptions.                                                                                                       //
  return this._backendCap.deleteGrain(ownerId, grainId)["catch"](function (err) {                                      // 86
    console.error("problem deleting grain " + grainId + ":", err.message);                                             // 87
    throw err;                                                                                                         // 88
  });                                                                                                                  // 89
};                                                                                                                     // 90
                                                                                                                       //
SandstormBackend.shouldRestartGrain = function (error, retryCount) {                                                   // 92
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      //
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          //
  // already gone through this cycle (should be zero for the first call).                                              //
                                                                                                                       //
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 97
};                                                                                                                     // 98
                                                                                                                       //
SandstormBackend.prototype.maybeRetryUseGrain = function (grainId, cb, retryCount, err) {                              // 100
  var _this = this;                                                                                                    // 101
  if (SandstormBackend.shouldRestartGrain(err, retryCount)) {                                                          // 102
    return inMeteor(function () {                                                                                      // 103
      return cb(_this.continueGrain(grainId).supervisor)["catch"](_this.maybeRetryUseGrain.bind(_this, grainId, cb, retryCount + 1));
    });                                                                                                                // 106
  } else {                                                                                                             // 107
    throw err;                                                                                                         // 108
  }                                                                                                                    // 109
};                                                                                                                     // 110
                                                                                                                       //
SandstormBackend.prototype.useGrain = function (grainId, cb) {                                                         // 112
  var _this2 = this;                                                                                                   // 112
                                                                                                                       //
  // This will open a grain for you, handling restarts if needed, and call the passed function with                    //
  // the supervisor capability as the only parameter. The callback must return a promise that used                     //
  // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                       //
  // This function returns the same promise that your callback returns.                                                //
  //                                                                                                                   //
  // This function is NOT expected to be run in a meteor context.                                                      //
                                                                                                                       //
  return inMeteor(function () {                                                                                        // 120
    return cb(_this2.continueGrain(grainId).supervisor)["catch"](_this2.maybeRetryUseGrain.bind(_this2, grainId, cb, 0));
  });                                                                                                                  // 123
};                                                                                                                     // 124
                                                                                                                       //
SandstormBackend.prototype.continueGrain = function (grainId) {                                                        // 126
  var grain = Grains.findOne(grainId);                                                                                 // 127
  if (!grain) {                                                                                                        // 128
    throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                            // 129
  }                                                                                                                    // 130
                                                                                                                       //
  // If a DevPackage with the same app ID is currently active, we let it override the installed                        //
  // package, so that the grain runs using the dev app.                                                                //
  var devPackage = DevPackages.findOne({ appId: grain.appId });                                                        // 134
  var isDev = void 0;                                                                                                  // 135
  var pkg = void 0;                                                                                                    // 136
  if (devPackage) {                                                                                                    // 137
    isDev = true;                                                                                                      // 138
    pkg = devPackage;                                                                                                  // 139
  } else {                                                                                                             // 140
    pkg = Packages.findOne(grain.packageId);                                                                           // 141
    if (!pkg) {                                                                                                        // 142
      throw new Meteor.Error(500, "Grain's package not installed", "Package ID: " + grain.packageId);                  // 143
    }                                                                                                                  // 145
  }                                                                                                                    // 146
                                                                                                                       //
  var manifest = pkg.manifest;                                                                                         // 148
  var packageId = pkg._id;                                                                                             // 149
                                                                                                                       //
  if (!("continueCommand" in manifest)) {                                                                              // 151
    throw new Meteor.Error(500, "Package manifest defines no continueCommand.", "Package ID: " + packageId);           // 152
  }                                                                                                                    // 154
                                                                                                                       //
  var result = this.startGrainInternal(packageId, grainId, grain.userId, manifest.continueCommand, false, isDev);      // 156
  result.packageSalt = isDev ? pkg._id : grain.packageSalt;                                                            // 158
  return result;                                                                                                       // 159
};                                                                                                                     // 160
                                                                                                                       //
SandstormBackend.prototype.startGrainInternal = function (packageId, grainId, ownerId, command, isNew, isDev) {        // 162
  // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                        //
  // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                     //
  // user) and `supervisor` (the supervisor capability).                                                               //
                                                                                                                       //
  if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                     // 167
    throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.\n" + "Please ask them to upgrade.");
  }                                                                                                                    // 171
                                                                                                                       //
  // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather                    //
  //   than just "argv".                                                                                               //
  if ("args" in command) {                                                                                             // 175
    if (!("argv" in command)) {                                                                                        // 176
      command.argv = command.args;                                                                                     // 177
    }                                                                                                                  // 178
                                                                                                                       //
    delete command.args;                                                                                               // 180
  }                                                                                                                    // 181
                                                                                                                       //
  if ("executablePath" in command) {                                                                                   // 183
    if (!("deprecatedExecutablePath" in command)) {                                                                    // 184
      command.deprecatedExecutablePath = command.executablePath;                                                       // 185
    }                                                                                                                  // 186
                                                                                                                       //
    delete command.executablePath;                                                                                     // 188
  }                                                                                                                    // 189
                                                                                                                       //
  return this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev);                              // 191
};                                                                                                                     // 192
                                                                                                                       //
SandstormBackend.prototype.updateLastActive = function (grainId, userId, identityId) {                                 // 194
  // Update the lastActive date on the grain, any relevant API tokens, and the user,                                   //
  // and also update the user's storage usage.                                                                         //
                                                                                                                       //
  var storagePromise = undefined;                                                                                      // 198
  if (Meteor.settings["public"].quotaEnabled) {                                                                        // 199
    storagePromise = globalBackend._backendCap.getUserStorageUsage(userId);                                            // 200
  }                                                                                                                    // 201
                                                                                                                       //
  var now = new Date();                                                                                                // 203
  if (Grains.update(grainId, { $set: { lastUsed: now } }) === 0) {                                                     // 204
    // Grain must have been deleted. Ignore.                                                                           //
    return;                                                                                                            // 206
  }                                                                                                                    // 207
                                                                                                                       //
  if (userId) {                                                                                                        // 209
    Meteor.users.update(userId, { $set: { lastActive: now } });                                                        // 210
  }                                                                                                                    // 211
                                                                                                                       //
  if (identityId) {                                                                                                    // 213
    Meteor.users.update({ _id: identityId }, { $set: { lastActive: now } });                                           // 214
    // Update any API tokens that match this user/grain pairing as well                                                //
    ApiTokens.update({ grainId: grainId, "owner.user.identityId": identityId }, { $set: { lastUsed: now } }, { multi: true });
  }                                                                                                                    // 219
                                                                                                                       //
  if (Meteor.settings["public"].quotaEnabled) {                                                                        // 221
    try {                                                                                                              // 222
      var ownerId = Grains.findOne(grainId).userId;                                                                    // 223
      var size = parseInt(waitPromise(storagePromise).size);                                                           // 224
      Meteor.users.update(ownerId, { $set: { storageUsage: size } });                                                  // 225
      // TODO(security): Consider actively killing grains if the user is excessively over quota?                       //
      //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                    //
    } catch (err) {                                                                                                    // 228
      if (err.kjType !== "unimplemented") {                                                                            // 229
        console.error("error getting user storage usage:", err.stack);                                                 // 230
      }                                                                                                                // 231
    }                                                                                                                  // 232
  }                                                                                                                    // 233
};                                                                                                                     // 234
                                                                                                                       //
function generateSessionId(grainId, userId, packageSalt, clientSalt) {                                                 // 236
  var sessionParts = [grainId, clientSalt];                                                                            // 237
  if (userId) {                                                                                                        // 238
    sessionParts.push(userId);                                                                                         // 239
  }                                                                                                                    // 240
                                                                                                                       //
  if (packageSalt) {                                                                                                   // 242
    sessionParts.push(packageSalt);                                                                                    // 243
  }                                                                                                                    // 244
                                                                                                                       //
  var sessionInput = sessionParts.join(":");                                                                           // 246
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 247
}                                                                                                                      // 248
                                                                                                                       //
SandstormBackend.prototype.openSessionInternal = function (grainId, userId, identityId, title, apiToken, cachedSalt) {
                                                                                                                       //
  // Start the grain if it is not running. This is an optimization: if we didn't start it here,                        //
  // it would start on the first request to the session host, but we'd like to get started before                      //
  // the round trip.                                                                                                   //
                                                                                                                       //
  var _continueGrain = this.continueGrain(grainId);                                                                    // 250
                                                                                                                       //
  var supervisor = _continueGrain.supervisor;                                                                          // 250
  var packageSalt = _continueGrain.packageSalt;                                                                        // 250
                                                                                                                       //
                                                                                                                       //
  this.updateLastActive(grainId, userId, identityId);                                                                  // 257
                                                                                                                       //
  cachedSalt = cachedSalt || Random.id(22);                                                                            // 259
  var sessionId = generateSessionId(grainId, userId, packageSalt, cachedSalt);                                         // 260
  var session = Sessions.findOne({ _id: sessionId });                                                                  // 261
  if (session) {                                                                                                       // 262
    // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                     //
    if (session.identityId && session.identityId !== identityId || session.grainId !== grainId) {                      // 264
      var e = new Meteor.Error(500, "Duplicate SessionId");                                                            // 266
      console.error(e);                                                                                                // 267
      throw e;                                                                                                         // 268
    } else {                                                                                                           // 269
      return {                                                                                                         // 270
        supervisor: supervisor,                                                                                        // 271
        methodResult: {                                                                                                // 272
          sessionId: session._id,                                                                                      // 273
          title: title,                                                                                                // 274
          grainId: grainId,                                                                                            // 275
          hostId: session.hostId,                                                                                      // 276
          tabId: session.tabId,                                                                                        // 277
          salt: cachedSalt                                                                                             // 278
        }                                                                                                              // 272
      };                                                                                                               // 270
    }                                                                                                                  // 281
  }                                                                                                                    // 282
                                                                                                                       //
  // TODO(someday): Allow caller to specify the parent session from which to inherit the tab ID, or                    //
  //   something.                                                                                                      //
                                                                                                                       //
  session = {                                                                                                          // 287
    _id: sessionId,                                                                                                    // 288
    grainId: grainId,                                                                                                  // 289
    hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                                  // 290
    tabId: Crypto.createHash("sha256").update("tab:").update(sessionId).digest("hex").slice(0, 32),                    // 291
    timestamp: new Date().getTime(),                                                                                   // 292
    hasLoaded: false                                                                                                   // 293
  };                                                                                                                   // 287
                                                                                                                       //
  if (userId) {                                                                                                        // 296
    session.identityId = identityId;                                                                                   // 297
    session.userId = userId;                                                                                           // 298
  } else if (apiToken) {                                                                                               // 299
    session.hashedToken = apiToken._id;                                                                                // 300
  } else {}                                                                                                            // 301
  // Must be old-style sharing, i.e. !grain.private.                                                                   //
  // jscs:ignore disallowEmptyBlocks                                                                                   //
                                                                                                                       //
  Sessions.insert(session);                                                                                            // 305
                                                                                                                       //
  return {                                                                                                             // 307
    supervisor: supervisor,                                                                                            // 308
    methodResult: {                                                                                                    // 309
      sessionId: session._id,                                                                                          // 310
      title: title,                                                                                                    // 311
      grainId: grainId,                                                                                                // 312
      hostId: session.hostId,                                                                                          // 313
      tabId: session.tabId,                                                                                            // 314
      salt: cachedSalt                                                                                                 // 315
    }                                                                                                                  // 309
  };                                                                                                                   // 307
};                                                                                                                     // 318
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}}}},{"extensions":[".js",".json"]});
require("./node_modules/meteor/sandstorm-backend/sandstorm-backend.js");

/* Exports */
if (typeof Package === 'undefined') Package = {};
(function (pkg, symbols) {
  for (var s in symbols)
    (s in pkg) || (pkg[s] = symbols[s]);
})(Package['sandstorm-backend'] = {}, {
  SandstormBackend: SandstormBackend
});

})();

//# sourceMappingURL=sandstorm-backend.js.map
